/* A faster variant of the Er's LLHanoi TH loopless algorithm
   with the remainder operator (%) replaced by faster bit operations
   performing left and right cyclic rotations of integers 1,2,3 */

#include <stdio.h>
#include <stdlib.h>

#define PR      (void)printf(
#define PE      (void)fprintf(stderr,

#define ALLO(x) { if((x = (int *)malloc((n+2) * sizeof(int))) == NULL) {\
                  PE #x " allocation failed!\n"); exit(1); } }


main(int argc, char *argv[])
/*========================*/
{
 int i, dir, *D, *s, n, j;

 n = atoi(argv[1]);
 j = n+1;
 ALLO(D)
 ALLO(s)

 for(i=0; i<=j; i++) {
   D[i] = 1; s[i] = i+1;
 }

 dir = n&1;

 for(;;) {
  i = s[0];
  if(i>n) break;
  j=D[i];
  /*
  PR"move disc %d from %d ", i, j);
  PR"to %d\n", D[i]=(i&1)==dir ? ~j>>(j&1)&3 : ++j>>(j>>2<<1));
  */
  D[i]=(i&1)==dir ? ~j>>(j&1)&3 : ++j>>(j>>2<<1);
  s[0] = 1;
  s[i-1] = s[i];
  s[i] = i+1;
 }
}

/*
Left cyclic rotation replaces j=1,2,3 by l=3,1,2
Right rotation j=1,2,3 by r=2,3,1
In the original Er's algorithm:
	l = (j-1)%3
	r = (j+1)%3

Here these are replaced by (for some compilers very much) faster
expressions involving bit operations:
	l = ~j>>(j&1)&3
	r = ++j>>(j>>2<<1)

For the left rotation there are e.g. these expressions giving
identical results:
	l = (~j>>1&3)-(~j&1)
	l = j-1|(j+6&6)>>1
	l = j-1+((j+6&6)>>1)
But they are slightly more complex (slower).

For the right rotation, alternatives are e.g.:
	r = ++j&3|j>>2
	r = j+1&3|j-1>>1
(in both cases, | can again be replaced by +).

Is there something simpler?

M. Kolar, December 1, 2000
*/
